package explorviz.hpc_monitoring.reader;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import com.lmax.disruptor.EventHandler;
import com.lmax.disruptor.RingBuffer;
import com.lmax.disruptor.dsl.Disruptor;

import explorviz.hpc_monitoring.disruptor.RecordEvent;
import explorviz.hpc_monitoring.filter.counting.CountingThroughputFilter;
import explorviz.hpc_monitoring.filter.reconstruction.TraceReconstructionFilter;
import explorviz.hpc_monitoring.record.HostApplicationMetadata;
import explorviz.hpc_monitoring.record.IRecord;
import explorviz.hpc_monitoring.record.events.normal.AfterFailedOperationEvent;
import explorviz.hpc_monitoring.record.events.normal.AfterOperationEvent;
import explorviz.hpc_monitoring.record.events.normal.BeforeOperationEvent;
import gnu.trove.map.hash.TIntObjectHashMap;

public final class TCPReader {
	private static final int MESSAGE_BUFFER_SIZE = 131072;
	private final int listeningPort;

	private ServerSocketChannel serversocket;
	private boolean active = true;

	private final RingBuffer<RecordEvent> ringBuffer;

	private final ByteBuffer buffer;

	private static final CountingThroughputFilter counter = new CountingThroughputFilter(
			"Records per second");

	private final static TIntObjectHashMap<String> stringRegistry = new TIntObjectHashMap<String>(
			64);

	private final static List<byte[]> waitingForStringMessages = new ArrayList<byte[]>(
			1024);
	private HostApplicationMetadata hostApplicationMetadata;

	public TCPReader(final int listeningPort,
			final EventHandler<RecordEvent> endReceiver)
			throws IllegalArgumentException {
		this.listeningPort = listeningPort;

		buffer = ByteBuffer.allocateDirect(MESSAGE_BUFFER_SIZE);

		final ExecutorService exec = Executors.newCachedThreadPool();
		final Disruptor<RecordEvent> disruptor = new Disruptor<RecordEvent>(
				RecordEvent.EVENT_FACTORY, 32768, exec);

		@SuppressWarnings("unchecked")
		final EventHandler<RecordEvent>[] eventHandlers = new EventHandler[1];
		eventHandlers[0] = new TraceReconstructionFilter(1 * 1000 * 1000,
				endReceiver);
		disruptor.handleEventsWith(eventHandlers);
		ringBuffer = disruptor.start();
	}

	public void read() {
		try {
			open();
			while (active) {
				// TODO only one connection!
				final SocketChannel socketChannel = serversocket.accept();
				while ((socketChannel.read(buffer)) != -1) {
					buffer.flip();
					messagesfromByteArray(buffer);
				}

				serversocket.close();
			}
		} catch (final IOException ex) {
			System.out.println("Error in read() " + ex.toString());
		} finally {
			try {
				serversocket.close();
			} catch (final IOException e) {
				System.out.println("Error in read()" + e.toString());
			}
		}
	}

	private void open() throws IOException {
		serversocket = ServerSocketChannel.open();
		serversocket.socket().bind(new InetSocketAddress(listeningPort));
		System.out.println("listening on port " + listeningPort);
	}

	private void putInRingBuffer(final IRecord message) {
		counter.inputObjects(message);
		// System.out.println(message.toString());
		// final long hiseq = ringBuffer.next();
		// final RecordEvent valueEvent = ringBuffer.get(hiseq);
		// valueEvent.setValue(message);
		// ringBuffer.publish(hiseq);
	}

	public void terminate(final boolean error) {
		System.out.println("Shutdown of TCPReader requested.");
		active = false;
	}

	public void messagesfromByteArray(final ByteBuffer buffer) {
		while (buffer.remaining() > 0) {
			buffer.mark();
			try {
				final byte clazzId = buffer.get();

				switch (clazzId) {
					case HostApplicationMetadata.CLAZZ_ID: {
						readInTraceMetadata(buffer);
						break;
					}
					case BeforeOperationEvent.CLAZZ_ID: {
						readInBeforeOperationEvent(buffer);
						break;
					}
					case AfterFailedOperationEvent.CLAZZ_ID: {
						readInAfterFailedOperationEvent(buffer);
						break;
					}
					case AfterOperationEvent.CLAZZ_ID: {
						readInAfterOperationEvent(buffer);
						break;
					}
					case 4: {
						final int mapId = buffer.getInt();
						final int stringLength = buffer.getInt();

						final byte[] stringByteArray = new byte[stringLength];

						buffer.get(stringByteArray);

						addToRegistry(mapId, new String(stringByteArray));
						break;
					}
					default: {
						System.out.println("unknown class id " + clazzId
								+ " at offset " + (buffer.position() - 4));
					}
				}
			} catch (final BufferUnderflowException e) {
				buffer.reset();
				buffer.compact();
				return;
			}
		}

		buffer.clear();
	}

	private final void readInTraceMetadata(final ByteBuffer buffer) {
		final int hostnameId = buffer.getInt();
		final int applicationId = buffer.getInt();

		final String hostname = getStringFromRegistry(hostnameId);
		final String application = getStringFromRegistry(applicationId);

		// if ((hostname != null) && (application != null)) {
		hostApplicationMetadata = new HostApplicationMetadata(hostname,
				application);
		// } else {
		// final byte[] message = new
		// byte[TraceMetadata.BYTE_LENGTH_WITH_CLAZZ_ID];
		// System.arraycopy(b, offset
		// - TraceMetadata.BYTE_LENGTH_WITH_CLAZZ_ID, message, 0,
		// TraceMetadata.BYTE_LENGTH_WITH_CLAZZ_ID);
		// putInWaitingMessages(message);
		// }
	}

	private final void readInBeforeOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);

		// if (operation != null) {
		putInRingBuffer(new BeforeOperationEvent(timestamp, traceId,
				orderIndex, operation));
		// } else {
		// final byte[] message = new
		// byte[BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
		// System.arraycopy(b, offset
		// - BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID, message,
		// 0, BeforeOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
		// putInWaitingMessages(message);
		// }
	}

	private final void readInAfterFailedOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();
		final int causeId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		final String cause = getStringFromRegistry(causeId);

		// if ((operation != null) && (cause != null)) {
		putInRingBuffer(new AfterFailedOperationEvent(timestamp, traceId,
				orderIndex, operation, cause));
		// } else {
		// final byte[] message = new
		// byte[AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
		// System.arraycopy(b, offset
		// - AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID,
		// message, 0,
		// AfterFailedOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
		// putInWaitingMessages(message);
		// }
	}

	private final void readInAfterOperationEvent(final ByteBuffer buffer) {
		final long timestamp = buffer.getLong();
		final long traceId = buffer.getLong();
		final int orderIndex = buffer.getInt();
		final int operationId = buffer.getInt();

		final String operation = getStringFromRegistry(operationId);
		// if (operation != null) {
		putInRingBuffer(new AfterOperationEvent(timestamp, traceId, orderIndex,
				operation));
		// } else {
		// final byte[] message = new
		// byte[AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID];
		// System.arraycopy(b, offset
		// - AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID, message,
		// 0, AfterOperationEvent.BYTE_LENGTH_WITH_CLAZZ_ID);
		// putInWaitingMessages(message);
		// }

	}

	private void putInWaitingMessages(final byte[] message) {
		waitingForStringMessages.add(message);
	}

	private void checkWaitingMessages() {
		final List<byte[]> localWaitingList = new ArrayList<byte[]>();
		for (final byte[] waitingMessage : waitingForStringMessages) {
			localWaitingList.add(waitingMessage);
		}
		waitingForStringMessages.clear();

		// for (final byte[] waitingMessage : localWaitingList) {
		// final int waitingMessageClazzId = UnsafeBits.getInt(waitingMessage,
		// 0);
		// switch (waitingMessageClazzId) {
		// case TraceMetadata.CLAZZ_ID:
		// readInTraceMetadata(waitingMessage, 4);
		// break;
		// case BeforeOperationEvent.CLAZZ_ID:
		// readInBeforeOperationEvent(waitingMessage, 4);
		// break;
		// case AfterFailedOperationEvent.CLAZZ_ID:
		// readInAfterFailedOperationEvent(waitingMessage, 4);
		// break;
		// case AfterOperationEvent.CLAZZ_ID:
		// readInAfterOperationEvent(waitingMessage, 4);
		// break;
		// default:
		// break;
		// }
		// }
	}

	public void addToRegistry(final int key, final String value) {
		stringRegistry.put(key, value);

		// System.out.println("put key " + key + " value " + value);

		checkWaitingMessages();
	}

	private String getStringFromRegistry(final int id) {
		return stringRegistry.get(id);
	}
}
